# Before `make install' is performed this script should be runnable with
# `make test'. After `make install' it should work as `perl test.pl'

######################### We start with some black magic to print on failure.

# Change 1..1 below to 1..last_test_to_print .
# (It may become useful if the test is moved to ./t subdirectory.)

BEGIN { $| = 1; print "1..7\n"; }
END {print "not ok 1\n" unless $loaded;}
use SWISH::Stemmer;
$loaded = 1;
print "ok 1\n";

######################### End of black magic.

# Insert your test code below (better if it prints "ok 13"
# (correspondingly "not ok 13") depending on the success of chunk 13
# of the test code):

my %testwords = (
    money   => 'monei',
    running     => 'run',
    monies      => 'moni',
    libraries   => 'librari',
    library	=> 'librari',
    Italian	=> 'italian',
    Italy       => 'itali',    
);

my $test = 2;

for my $word ( sort keys %testwords ) {
    my $stem = SwishStem( $word );

    print "$stem = SwishStem( $word ) ";
    print $stem eq $testwords{ $word } ? "ok $test\n" : "not ok $test.  Should be '$testwords{ $word }'\n";
    $test++;
}
    

   

